(* Pierre Castéran, LaBRI and University of Bordeaux *)

Require Import Hydra_Games Epsilon0 KS Relation_Operators WfVariant.

(** We generalize the results of libraries Omega_Small and Omega2_Small:
   We prove the no ordinal strictly less than epsilon0 can be used as a variant
   for proving the termination of all hydra battles.

   In this proof, we use the so-called "Ketonen-Solovay machinery" for building
a proof that shares the same structure as for the libraries above, but is much 
longer 
 *)




Section Epsilon0_needed.
  Open Scope cnf_scope.
  (** Let us consider some ordinal less than epsilon0 *)

  Variable alphamax: T1.
  Hypothesis alphamax_nf : nf alphamax.

  Let t : Type := {alpha : T1 | alpha < alphamax}.
  
  (** Let us consider a variant from type [Hydra] into alphamax *)
  Variable m : Hydra -> T1.
  Hypothesis m_bounded : forall h, m h < alphamax.
  
  Context (V : WfVariant round LT m).
  Hint Resolve (decr V).

  Lemma m_variant_trans : forall h h', h -+-> h' -> m h' < m h.
  Proof.
    induction 1; auto.
    apply LT_trans with (m y);  auto.
  Qed.

  (** Let us transform any ordinal notation into an hydra *)

  Fixpoint iota (alpha : T1) : Hydra :=
    match alpha with
      | zero => head
      | ocons gamma n beta => node (hcons_mult (iota gamma) (S n)
                                               (iotas beta))
    end 
  with iotas (alpha : T1) :  Hydrae :=
         match alpha with
             | zero => hnil
             | ocons alpha0 n beta  => hcons_mult (iota alpha0) (S n)
                                                  (iotas beta)
         end.

  (**  We now prove  a lot of technical lemmas that relate Hydras and ordinals. *)
  
  
  Fixpoint hy_app (s1 s2: Hydrae) : Hydrae :=
    match s1 with
        hnil => s2
      | hcons h1 s1' => hcons h1 (hy_app s1' s2)
    end.

  Lemma hy_app_assoc : forall s1 s2 s3,
                         hy_app s1 (hy_app s2 s3) = hy_app (hy_app s1 s2) s3.
  Proof.
    induction s1.
    - reflexivity.
    - intros s2 s3; cbn; now rewrite IHs1.
  Qed.

  Lemma S0_app : forall s1 s2 s3, S0 s2 s3 ->
                                  S0 (hy_app s1 s2)
                                     (hy_app s1 s3).
  Proof.
    induction s1.
    - cbn; auto.  
    - intros; cbn;  constructor;auto.
  Qed.


  Lemma R1_app : forall s1 s2 s3,
                   R1 (node s2) (node s3) ->
                   R1 (node (hy_app s1 s2)) (node (hy_app s1 s3)).
  Proof.
    constructor.
    inversion  H; now apply S0_app.
  Qed.

  Lemma S1_app n : forall s1 s2 s3,
                     S1 n s2 s3 -> S1 n (hy_app s1 s2)
                                      (hy_app s1 s3).
  Proof.
    induction s1; simpl hy_app.
    - trivial.  
    - intros s2 s3 H; constructor; auto.
  Qed.


  Lemma S2_app : forall  n  s1 s2 ,
                   S2 n s1 s2 ->
                   forall l,  S2 n (hy_app l s1) (hy_app l s2).
  Proof.
    intros n s1 s2 h;
    eapply  S2_ind2 with
    (P := fun h1 h2 _ => match h1, h2 with
                             node s1, node s2 =>
                             forall l, R2 n (node (hy_app l s1))
                                          (node (hy_app l s2)) 
                         end).
    - intros; constructor.
      apply S1_app; auto.
      intros;  eauto. 
    - constructor 2.
      induction l.
      + cbn; auto.
      + cbn; now constructor 2.
    -  intros h0 h' s r  H l;  destruct h0, h';  induction l.   
       + simpl;auto.
       +  simpl;auto; constructor 2;auto.
    - auto.
    -  eexact h.
  Qed.


  Lemma R2_app n : forall  s1 s2 ,
                     R2 n (node s1) (node  s2) ->
                     forall l,  R2 n (node (hy_app l s1))
                                   (node (hy_app l s2)).
  Proof.
    intros s1 s2 H;  inversion  H.
    - intros; constructor.
      apply S1_app; auto.
    -  constructor 2; now apply S2_app.
  Qed.


  Lemma succ_rw1 : forall alpha n beta, alpha <> zero ->
                                        succ (ocons alpha n beta)=
                                        ocons alpha n (succ beta).
  Proof.
    induction alpha.
    - destruct 1;auto.
    - cbn; auto. 
  Qed.

  Lemma iota_iotas : forall alpha, nf alpha -> node (iotas alpha) =
                                               iota alpha .
  Proof. 
    induction alpha.
    - reflexivity.
    - intros; now cbn. 
  Qed.

  Lemma hcons_mult_app : forall h n s s',
                         hcons_mult h n (hy_app s s') =
                         hy_app (hcons_mult h n s) s'.
  Proof.
    induction n.
    - now cbn. 
    - cbn; intros; now rewrite IHn.
  Qed.


  Lemma iotas_succ : forall alpha, nf alpha ->
                                  iotas (succ alpha) =
                                  hy_app (iotas alpha) (hcons head hnil).
  Proof.
    induction  alpha.
    - cbn; trivial.
    - destruct alpha1; intros.
      + T1_inversion H.
        subst alpha2.
        replace (ocons zero n zero) with (FS  n).
        rewrite succ_compat.
        cbn;  clear  H;  induction n;simpl;auto.
        repeat f_equal.
        injection IHn; auto.
        reflexivity.
      +   rewrite succ_rw1; simpl hy_app.
          simpl iotas; repeat f_equal; rewrite <- hcons_mult_app;
          rewrite IHalpha2; auto. 
          eauto with T1.
          discriminate.
  Qed.

  Lemma hy_app_l_nil : forall l, hy_app l hnil = l.
  Proof.
    induction l;cbn; auto.
    rewrite IHl;auto.
  Qed.
  
  Lemma iota_succ_R1 : forall  o, nf o -> R1 (iota (succ o)) (iota o).
  Proof.
    intros; repeat rewrite <- iota_iotas; auto.
    rewrite iotas_succ; auto. 
    constructor.  
    replace  (iotas o) with (hy_app (iotas o) hnil) at 2.
    -  apply S0_app.
       constructor.
    -  now rewrite hy_app_l_nil.
    -  now apply succ_nf.
  Qed.


  Lemma iota_succ_round : forall  o, nf o -> iota (succ o) -1-> iota o.
  Proof.
    intros; exists 0; left;  now apply iota_succ_R1.
  Qed.

  Lemma iota_rw1 :
  forall i alpha,  nf alpha -> is_limit alpha = true ->
                   iota (dS i (ocons alpha 0 zero)) =
                   hyd1 (iota (dS i alpha)).
  Proof.                        
    intros i alpha H H0; rewrite  dS_lim1; auto.
  Qed.


  Lemma iota_rw2 :
    forall i n alpha,  nf alpha -> is_limit alpha = true ->
                       iota (dS i (ocons alpha (S n) zero)) =
                       node (hcons_mult (iota alpha) (S n) 
                                        (hcons
                                           (iota (dS i alpha)) hnil)).
  Proof.                        
    intros i n alpha H H0; rewrite  dS_lim2; auto.
  Qed.


  Lemma iota_rw3 :
    forall i  alpha,  nf alpha -> 
                      iota (dS i (ocons (succ alpha) 0 zero)) =
                      node (hcons_mult (iota alpha) (S i) hnil).
  Proof.                        
    intros i  alpha H; rewrite  dS_succ_eqn1; auto.
  Qed.

  Lemma iota_rw4 :
    forall i n  alpha ,  nf alpha -> 
                         iota (dS i (ocons (succ alpha) (S n) zero)) =
                         node (hcons_mult (iota (succ alpha)) (S n) 
                                          (hcons_mult (iota alpha) (S i) hnil)).
  Proof. intros i  n alpha H;  rewrite  dS_succ_eqn2; auto. Qed.


  Lemma iota_tail :
    forall i   alpha n beta,
      nf (ocons alpha n beta) ->
      beta <> zero ->
      iota (dS i (ocons alpha n beta)) =
      node (hcons_mult (iota alpha) (S n) (iotas (dS i beta)) ).
  Proof.                                  
    intros i   alpha n beta H H0; rewrite  dS_tail; auto.
  Qed.


  Lemma R1_hcons : forall h s1 s2, R1 (node s1) (node s2) ->
                                   R1 (node (hcons h s1)) (node (hcons h s2)).
  Proof. inversion 1;split;now right. Qed.

  Lemma R1_hcons_mult  : forall n h s1 s2,
                           R1 (node s1) (node s2) ->
                           R1 (node (hcons_mult  h n s1))
                              (node (hcons_mult  h n s2)).
  Proof. inversion 1; split; now apply hcons_mult_S0. Qed.

  Lemma R1_R2 : forall h h', R1 h h' -> R2 0 (hyd1 h) (hyd1  h').
  Proof.
    intros; destruct h, h'; left;
    change (hcons (node h0) hnil) with (hcons_mult (node h0) 1 hnil);
    now left.
  Qed.

  (** The sequence s contains a head *)
  
  Inductive mem_head : Hydrae -> Prop :=
    hh_1 : forall s, mem_head (hcons head s)
  | hh_2 : forall h s, mem_head s -> mem_head (hcons h s).


  Lemma S0_mem_head : forall s s', S0 s s' -> mem_head s.
  Proof.
    induction 1.
    -  left.
    -  now right.
  Qed.


  Lemma mem_head_mult_inv : forall n h s, mem_head (hcons_mult h n s) ->
                                          h = head \/ mem_head s.
  Proof. 
    induction n; cbn ; auto. 
    inversion 1;  auto.
  Qed. 


  Lemma R1_mem_head : forall l l', R1 (node l) (node l') -> mem_head l.
  Proof.
    inversion 1;   eapply S0_mem_head; eauto.
  Qed.

  Lemma limit_no_head : forall alpha, nf alpha -> is_limit alpha = true ->
                                      ~ mem_head (iotas alpha).
  Proof. 
    induction alpha.
    - discriminate.
    - intros; simpl;  destruct (is_limit_cases H).
      +  auto.
      +  destruct a;  subst; intro H2;  inversion H2.
         destruct alpha1.
         *    destruct H1;auto.
         *    discriminate H4.
         *  destruct  (mem_head_mult_inv _ _ _ H4).
            destruct alpha1.
            { destruct H1;   auto.  }
            discriminate H6.
            inversion H6.
      +  intro H7;
        change
          (hcons (iota alpha1) (hcons_mult (iota alpha1) n (iotas alpha2)))
        with   (hcons_mult (iota alpha1) (S n) (iotas alpha2)) in H7.
         destruct  (mem_head_mult_inv _ _ _ H7).
         *    destruct alpha1.
              destruct a.
              destruct H2;auto.
              discriminate.
         *   destruct IHalpha2;  eauto with T1.
             tauto. 
  Qed.


  Lemma limit_no_R1 : forall alpha, nf alpha ->
                                    is_limit alpha = true ->
                                    forall h', ~ R1 (iota alpha) h'.
  Proof.
    intros alpha H H0 h' H1; generalize ( limit_no_head _ H H0).
    intro H2; case_eq  (iota alpha); case_eq h';intros.
    subst h';rewrite <- iota_iotas in H1; auto. 
    generalize (R1_mem_head _ _ H1).
    tauto.
  Qed.

  (**  The following (long! ) proof relates the rounds of the hydra game
      with the canonical sequences *)
  
  Lemma d_S_iota : forall i o, nf o -> o <> zero -> (iota o) -1-> (iota (dS i o)).
  Proof.
    induction o.
    { (* trivial case o = zero *)
      now destruct 2.
    }
    intros  H H0; destruct (zero_succ_limit_dec H).
    - destruct s.
      {  discriminate.  }
      clear H0; destruct (is_limit_cases H i0).
      + destruct a; subst o2.
        destruct n.
      * (* n = 0 *)
        destruct (zero_succ_limit_dec (alpha := o1)).
       { inversion H;auto. }
       { destruct s.
         { (* o1 = zero *)   destruct H0. auto. }
         { (* o1 limit *)
           rewrite dS_lim1; eauto with T1.
           cbn; destruct IHo1; eauto with T1.
           inversion H;auto. 
           destruct H1.
           - exists 0;   right.
             left; left; auto. 
          - exists x; right. 
           right; left;  auto. 
         }
        }
         (* o1 is the successor of gamma *) 
         destruct s as [gamma [Hnf eo1]]; subst o1.
         rewrite iota_rw3; eauto with T1.
         simpl (iota (ocons (succ gamma) 0 zero));
         exists i; constructor 2.
         constructor .     
         left;  apply iota_succ_R1;  auto. 
      *  destruct (zero_succ_limit_dec (alpha := o1)).
        { inversion H;auto. }
        { destruct s.
         { (* o1 = zero *)   destruct H0; auto. }
         { (* o1 limit *)
         rewrite iota_rw2; eauto with T1.
         simpl  (iota (ocons o1 (S n) zero)).
         simpl (hcons_mult (iota o1) (S n) (hcons (iota (dS i o1)) hnil)).
         destruct IHo1; eauto with T1.
         replace (hcons (iota o1)
                        (hcons (iota o1) (hcons_mult (iota o1) n hnil)))
         with (hcons (iota o1) (hcons_mult (iota o1) n (hcons (iota o1) hnil))).
         destruct H1.
         { exists 0; right.
           left; right; apply hcons_mult_S1.
           change (hcons (iota (dS i o1)) hnil) with
           (hcons_mult (iota (dS i o1)) 1 hnil); now left. 
         }
         {
           exists x; right;  right; right;  apply hcons_mult_S2.
           now left.
         }
         now rewrite hcons_mult_comm; auto.
         }
        }
        destruct s as [gamma [Hnf eo1]]; subst o1.
        rewrite iota_rw4; auto.
        simpl (iota (ocons (succ gamma) (S n) zero)); destruct IHo1.
        apply succ_nf;auto. 
        auto.
        rewrite dS_succ in H1.
         exists i; right. 
         simpl (hcons_mult (iota (succ gamma)) (S n)
                           (hcons_mult (iota gamma) (S i) hnil)).
         left; right; rewrite  <- hcons_mult_comm.
         apply hcons_mult_S1.
         left;  apply iota_succ_R1; auto.
         auto. 
      +  rewrite iota_tail; eauto with T1.
         * destruct a; simpl; destruct IHo2.
           eauto with T1. 
           intro H3; subst o2; discriminate.   
           destruct H2. 
           { 
             eexists; left.
             rewrite <- iota_iotas in H2.
             rewrite <- iota_iotas in H2.
             inversion H2.
             split.
             right;  apply hcons_mult_S0.
             inversion H3;  auto.
             apply nf_dS.
             eauto with T1.
             eauto with T1.
           }
           exists x.
           replace (hcons (iota o1) (hcons_mult (iota o1) n (iotas o2))) with
           (hy_app (hcons_mult (iota o1) (S n) hnil) (iotas o2)).
           replace (hcons (iota o1) (hcons_mult (iota o1) n (iotas (dS i o2)))) with
           (hy_app (hcons_mult (iota o1) (S n) hnil) (iotas (dS i o2))).
           right; apply R2_app.
           rewrite <- iota_iotas in H2.
           rewrite <- iota_iotas in H2.
           auto.
           apply nf_dS; eauto with T1. 
           eauto with T1. 
        cbn; f_equal;  clear H2 H1 H0  i0 H IHo1.
        induction n;simpl;auto.
        f_equal;  auto.
        clear H2 H1 H0 i0 H IHo1; induction n;simpl;auto.
        f_equal.
        auto.
         *  intro ; subst o2; destruct a; discriminate.
    - destruct s.
      destruct a.
      rewrite H2.
      rewrite dS_succ; auto. 
      apply iota_succ_round; auto. 
      Unshelve.
      exact 0.
  Qed.

  (** We are now able to find a contradiction in the hypotheses *)
  
  Let h := iota alphamax.
  Let beta_h := m h.
  Let h' := iota beta_h.
 
  Remark R1 : beta_h < alphamax.
  Proof.
    unfold beta_h; auto.
  Qed.

 (** We apply a result from Ketonen and Solovay.
    beta_h is reachable from alphamax by a sequence of
    dS n operations *)
  
  Remark R2 : {n:nat | dS_plus n alphamax beta_h}.
   apply L2_6_1; auto. 
   apply m_bounded.
  Qed.

  Remark R3 : alphamax <> zero.
  Proof. 
    intro H; subst;  specialize (m_bounded (hyd1 head)).
    inversion m_bounded as [H H0].
    destruct H0; T1_inversion H0.
  Qed.

  Remark R4 : forall n alpha beta, nf alpha -> nf beta -> alpha <> zero ->
                                   dS_rel n alpha beta ->
                                   iota alpha -+-> iota beta.
  Proof.
     inversion 4; subst; left; apply d_S_iota; auto.
  Qed.   

  Remark R5 : forall n alpha beta, nf alpha -> nf beta -> alpha <> zero ->
                                   dS_plus n alpha beta ->
                                   iota alpha -+-> iota beta.
  Proof.
    induction 4.
    -  apply R4 with n;auto.
    -  destruct (T1_eq_dec y zero).
       +  subst;  assert (z=zero).
          {
            clear H2 H1 H0 IHclos_trans_1n.
            assert
              (forall x a,
                 clos_trans_1n T1 (dS_rel n) x a -> x = zero -> a = zero).
            { induction 1.
              - intro ; subst; inversion H0;  reflexivity.
              -  intros; apply IHclos_trans_1n.
                 subst; inversion H0; reflexivity.
            }
            eapply H0; eauto.
          }
          subst z; left; inversion H2.
          rewrite H5;  apply d_S_iota;   auto. 
        + constructor 2 with (iota y).
          * inversion H2; apply d_S_iota;  auto. 
          * apply IHclos_trans_1n; auto. 
            inversion H2.
            apply nf_dS; auto. 
  Qed.

 Corollary  C5 : forall alpha beta,  alpha < beta ->
                                     iota beta -+-> iota alpha.
 Proof. 
  intros alpha beta H.
  assert (H0: nf beta) by eauto with T1.
  destruct (L2_6_1 H0 H) as [n H1].
  eapply R5; eauto with T1.
  intro;subst beta; apply (not_LT_zero H).
 Qed. 


  Remark R6 : h -+-> h'.
  Proof. 
    unfold h, h';  destruct R2.  apply R5 with x; auto.
    destruct (m_bounded h).
    - apply H. 
    - apply R3.
 Qed.

  Corollary mh'_lt_mh : m h' < m h. 
  Proof. apply m_variant_trans; apply R6. Qed.

  Remark R8 : forall h, nf (m h).
  Proof.
    intro h0;now destruct (m_bounded h0).
  Qed.

  (** With a transfinite induction,
     we prove the second inequality for getting a contradiction.

     In the case where alpha is a limit ordinal, we use the fact that
     alpha is the least upper bound of all the (d i alpha). 
 
 *)

  Lemma m_ge : forall alpha, nf alpha -> alpha <= m (iota alpha).
  Proof.
   intros alpha ; transfinite_induction alpha.
   clear alpha; intros alpha Hrec H.
     destruct (T1_eq_dec alpha zero).
      + subst;  split;auto with T1.
        split.
        auto with T1.
        apply R8.
      +   destruct (zero_succ_limit_dec H).
          * destruct s.
            { subst ; now destruct n. }
            { eapply limit_strict_lub with (s := fun i => dS i alpha).
              apply dS_limit; auto. 
              intro i0;  assert (iota alpha -1-> iota (dS i0 alpha)). 
              {   apply d_S_iota;  auto. }
              apply LE_LT_trans with (m (iota (dS i0 alpha))); auto.
              { apply Hrec; auto. 
                apply nf_dS;  auto.
                destruct (dS_limit H); auto. 
                destruct (H1 i0).
                tauto.
                eapply nf_dS;auto.
              }  
            }
          *  destruct s.
             destruct a; subst.
             assert (x < m (iota (succ x))).
             { 
               apply LE_LT_trans with (m (iota x)).
               apply Hrec; auto. 
               apply lt_succ; auto. 
               apply (decr V).
               apply     iota_succ_round; auto.
             }
             split; auto. 
             split; auto. 
             apply  lt_succ_le; auto. 
             apply R8.
             destruct H1; tauto.
             apply R8; auto. 
  Qed.

  Fact F : m h <  m h.
  Proof. 
    apply LE_LT_trans with (m h').
    - unfold h, h', beta_h;  unfold h;  apply m_ge;  apply R8.
    -  apply mh'_lt_mh.
  Qed. 

  Theorem epsilon0_needed : False.
  Proof.
    apply (LT_irrefl F).
  Qed.

End Epsilon0_needed.




