(* Pierre Castéran, LaBRI and University of Bordeaux *)

Require Import Hydra_Games Epsilon0 KS Relation_Operators  RelationClasses
WfVariant TooSimpleVariant.

(** We generalize the results of libraries Omega_Small and Omega2_Small:
   We prove the no ordinal strictly less than epsilon0 can be used as a variant
   for proving the termination of all hydra battles.

   In this proof, we use the so-called "Ketonen-Solovay machinery" for building
a proof that shares the same structure as for the libraries above, but is much 
longer 
 *)




Section Epsilon0_needed.
  Open Scope cnf_scope.
  
  (** Let us consider some ordinal less than epsilon0 *)

  Variable alphamax: T1.
  Hypothesis alphamax_nf : nf alphamax.

  Let t : Type := {alpha : T1 | alpha < alphamax}.

  (* strict order on t *)

  Let Lt (alpha beta : t) :=
    match alpha, beta with
        exist _ t1 _, exist _ t2 _ => lt t1 t2
    end.

  
  Instance Lt_irr : Irreflexive Lt.
  Proof.
    intros x H;   unfold Lt in  H;   destruct x;
    now apply (@lt_irrefl x).
  Qed.

  Instance Lt_trans : Transitive Lt.
  Proof.
    red; destruct x,y,z; unfold Lt; intros; eapply lt_trans; eauto.
  Qed.


  Section Assumed_Variant.
    
    (** Let us consider a variant from type Hydra into the set 
     of ordinals stricty less than alphamax *)

    Variable m : Hydra -> t.

    (* conversion of m to type T1 *)
    
    Let m_extract h := match m h with exist _ t _ => t end.    

    Lemma  m_bounded : forall h, lt (m_extract h) alphamax.
    Proof.     
      unfold m_extract;  intro h; now destruct (m h) as [x [_ [H2 _]]].  
    Qed.


    Context (V : WfVariant round Lt m).
    Hint Resolve (decr V).

    Lemma m_variant_Lt : forall h h', h -+-> h' -> Lt (m h') (m h).
    Proof. 
     induction 1 as [| h h' h'' H];auto.
     - transitivity (m h'); auto.
    Qed.

    (** Let us transform any ordinal notation into an hydra *)

    Fixpoint iota (alpha : T1) : Hydra :=
      match alpha with
        | zero => head
        | ocons gamma n beta => node (hcons_mult (iota gamma) (S n)
                                                 (iotas beta))
      end 
    with iotas (alpha : T1) :  Hydrae :=
           match alpha with
             | zero => hnil
             | ocons alpha0 n beta  => hcons_mult (iota alpha0) (S n)
                                                  (iotas beta)
           end.

    Definition Iota (alpha: t) : Hydra := iota (proj1_sig alpha).



    (**  We now prove  a lot of technical lemmas that relate 
       Hydras and ordinals. *)
    
  
   
    

    
    Lemma succ_rw1 : forall alpha n beta, alpha <> zero ->
                                          succ (ocons alpha n beta)=
                                          ocons alpha n (succ beta).
    Proof.
      induction alpha.
      - destruct 1;auto.
      - cbn; auto. 
    Qed.

    Lemma iota_iotas : forall alpha, nf alpha -> node (iotas alpha) =
                                                 iota alpha .
    Proof. 
      induction alpha.
      - reflexivity.
      - intros; now cbn. 
    Qed.

    Lemma hcons_mult_app : forall h n s s',
                             hcons_mult h n (hy_app s s') =
                             hy_app (hcons_mult h n s) s'.
    Proof.
      induction n.
      - now cbn. 
      - cbn; intros; now rewrite IHn.
    Qed.


    Lemma iotas_succ : forall alpha, nf alpha ->
                                     iotas (succ alpha) =
                                     hy_app (iotas alpha) (hcons head hnil).
    Proof.
      induction  alpha.
      - cbn; trivial.
      - destruct alpha1; intros.
        + T1_inversion H.
          subst alpha2.
          replace (ocons zero n zero) with (FS  n).
          rewrite succ_compat.
          cbn;  clear  H;  induction n;simpl;auto.
          repeat f_equal.
          injection IHn; auto.
          reflexivity.
        +   rewrite succ_rw1; simpl hy_app.
            simpl iotas; repeat f_equal; rewrite <- hcons_mult_app;
            rewrite IHalpha2; auto. 
            eauto with T1.
            discriminate.
    Qed.

    Lemma hy_app_l_nil : forall l, hy_app l hnil = l.
    Proof.
      induction l;cbn; auto.
      rewrite IHl;auto.
    Qed.
    
    Lemma iota_succ_R1 : forall  o, nf o -> R1 (iota (succ o)) (iota o).
    Proof.
      intros; repeat rewrite <- iota_iotas; auto.
      rewrite iotas_succ; auto. 
      constructor.  
      replace  (iotas o) with (hy_app (iotas o) hnil) at 2.
      -  apply S0_app.
         constructor.
      -  now rewrite hy_app_l_nil.
      -  now apply succ_nf.
    Qed.


    Lemma iota_succ_round : forall  o, nf o -> iota (succ o) -1-> iota o.
    Proof.
      intros; exists 0; left;  now apply iota_succ_R1.
    Qed.

    Lemma iota_rw1 :
      forall i alpha,  nf alpha -> is_limit alpha = true ->
                       iota (dS i (ocons alpha 0 zero)) =
                       hyd1 (iota (dS i alpha)).
    Proof.                        
      intros i alpha H H0; rewrite  dS_lim1; auto.
    Qed.


    Lemma iota_rw2 :
      forall i n alpha,  nf alpha -> is_limit alpha = true ->
                         iota (dS i (ocons alpha (S n) zero)) =
                         node (hcons_mult (iota alpha) (S n) 
                                          (hcons
                                             (iota (dS i alpha)) hnil)).
    Proof.                        
      intros i n alpha H H0; rewrite  dS_lim2; auto.
    Qed.


    Lemma iota_rw3 :
      forall i  alpha,  nf alpha -> 
                        iota (dS i (ocons (succ alpha) 0 zero)) =
                        node (hcons_mult (iota alpha) (S i) hnil).
    Proof.                        
      intros i  alpha H; rewrite  dS_succ_eqn1; auto.
    Qed.

    Lemma iota_rw4 :
      forall i n  alpha ,  nf alpha -> 
                           iota (dS i (ocons (succ alpha) (S n) zero)) =
                           node (hcons_mult (iota (succ alpha)) (S n) 
                                            (hcons_mult (iota alpha) (S i) hnil)).
    Proof. intros i  n alpha H;  rewrite  dS_succ_eqn2; auto. Qed.


    Lemma iota_tail :
      forall i   alpha n beta,
        nf (ocons alpha n beta) ->
        beta <> zero ->
        iota (dS i (ocons alpha n beta)) =
        node (hcons_mult (iota alpha) (S n) (iotas (dS i beta)) ).
    Proof.                                  
      intros i   alpha n beta H H0; rewrite  dS_tail; auto.
    Qed.


    Lemma R1_hcons : forall h s1 s2, R1 (node s1) (node s2) ->
                                     R1 (node (hcons h s1)) (node (hcons h s2)).
    Proof. inversion 1;split;now right. Qed.

    Lemma R1_hcons_mult  : forall n h s1 s2,
                             R1 (node s1) (node s2) ->
                             R1 (node (hcons_mult  h n s1))
                                (node (hcons_mult  h n s2)).
    Proof. inversion 1; split; now apply hcons_mult_S0. Qed.

    Lemma R1_R2 : forall h h', R1 h h' -> R2 0 (hyd1 h) (hyd1  h').
    Proof.
      intros; destruct h, h'; left;
      change (hcons (node h0) hnil) with (hcons_mult (node h0) 1 hnil);
      now left.
    Qed.

    (** The sequence s contains a head *)
    
    Inductive mem_head : Hydrae -> Prop :=
      hh_1 : forall s, mem_head (hcons head s)
    | hh_2 : forall h s, mem_head s -> mem_head (hcons h s).


    Lemma S0_mem_head : forall s s', S0 s s' -> mem_head s.
    Proof.
      induction 1.
      -  left.
      -  now right.
    Qed.


    Lemma mem_head_mult_inv : forall n h s, mem_head (hcons_mult h n s) ->
                                            h = head \/ mem_head s.
    Proof. 
      induction n; cbn ; auto. 
      inversion 1;  auto.
    Qed. 


    Lemma R1_mem_head : forall l l', R1 (node l) (node l') -> mem_head l.
    Proof.
      inversion 1;   eapply S0_mem_head; eauto.
    Qed.

    Lemma limit_no_head : forall alpha, nf alpha -> is_limit alpha = true ->
                                        ~ mem_head (iotas alpha).
    Proof. 
      induction alpha.
      - discriminate.
      - intros; simpl;  destruct (is_limit_cases H).
        +  auto.
        +  destruct a;  subst; intro H2;  inversion H2.
           destruct alpha1.
           *    destruct H1;auto.
           *    discriminate H4.
           *  destruct  (mem_head_mult_inv _ _ _ H4).
              destruct alpha1.
              { destruct H1;   auto.  }
              discriminate H6.
              inversion H6.
        +  intro H7;
          change
            (hcons (iota alpha1) (hcons_mult (iota alpha1) n (iotas alpha2)))
          with   (hcons_mult (iota alpha1) (S n) (iotas alpha2)) in H7.
           destruct  (mem_head_mult_inv _ _ _ H7).
           *    destruct alpha1.
                destruct a.
                destruct H2;auto.
                discriminate.
           *   destruct IHalpha2;  eauto with T1.
               tauto. 
    Qed.


    Lemma limit_no_R1 : forall alpha, nf alpha ->
                                      is_limit alpha = true ->
                                      forall h', ~ R1 (iota alpha) h'.
    Proof.
      intros alpha H H0 h' H1; generalize ( limit_no_head _ H H0).
      intro H2; case_eq  (iota alpha); case_eq h';intros.
      subst h';rewrite <- iota_iotas in H1; auto. 
      generalize (R1_mem_head _ _ H1).
      tauto.
    Qed.

    (**  The following (long! ) proof relates the rounds of the hydra game
      with the canonical sequences *)
    
    Lemma d_S_iota : forall i alpha, nf alpha->  alpha <> zero ->
                                 iota alpha -1-> iota (dS i alpha).
    Proof.
      induction alpha.
      { (* trivial case alpha = zero *)
        now destruct 2.
      }
      intros  H H0; destruct (zero_succ_limit_dec H).
      - destruct s.
        {  discriminate.  }
        clear H0; destruct (is_limit_cases H i0).
        + destruct a; subst alpha2.
          destruct n.
          * (* n = 0 *)
            destruct (zero_succ_limit_dec (alpha := alpha1)).
            { inversion H;auto. }
            { destruct s.
              { (* o1 = zero *)   destruct H0. auto. }
              { (* o1 limit *)
                rewrite dS_lim1; eauto with T1.
                cbn; destruct IHalpha1; eauto with T1.
                inversion H;auto. 
                destruct H1.
                - exists 0;   right.
                  left; left; auto. 
                - exists x; right. 
                  right; left;  auto. 
              }
            }
            (* alpha1 is the successor of gamma *) 
            destruct s as [gamma [Hnf eo1]]; subst alpha1.
            rewrite iota_rw3; eauto with T1.
            simpl (iota (ocons (succ gamma) 0 zero));
              exists i; constructor 2.
            constructor .     
            left;  apply iota_succ_R1;  auto. 
          *  destruct (zero_succ_limit_dec (alpha := alpha1)).
             { inversion H;auto. }
             { destruct s.
               { (* alpha1 = zero *)   destruct H0; auto. }
               { (* alpha1 limit *)
                 rewrite iota_rw2; eauto with T1.
                 simpl  (iota (ocons alpha1 (S n) zero)).
                 simpl (hcons_mult (iota alpha1) (S n)
                                   (hcons (iota (dS i alpha1)) hnil)).
                 destruct IHalpha1; eauto with T1.
                 replace (hcons (iota alpha1)
                                (hcons (iota alpha1)
                                       (hcons_mult (iota alpha1) n hnil)))
                 with (hcons (iota alpha1)
                             (hcons_mult (iota alpha1) n (hcons (iota alpha1) hnil))).
                 destruct H1.
                 { exists 0; right.
                   left; right; apply hcons_mult_S1.
                   change (hcons (iota (dS i alpha1)) hnil) with
                   (hcons_mult (iota (dS i alpha1)) 1 hnil); now left. 
                 }
                 {
                   exists x; right;  right; right;  apply hcons_mult_S2.
                   now left.
                 }
                 now rewrite hcons_mult_comm; auto.
               }
             }
             destruct s as [gamma [Hnf eo1]]; subst alpha1.
             rewrite iota_rw4; auto.
             simpl (iota (ocons (succ gamma) (S n) zero)); destruct IHalpha1.
             apply succ_nf;auto. 
             auto.
             rewrite dS_succ in H1.
             exists i; right. 
             simpl (hcons_mult (iota (succ gamma)) (S n)
                               (hcons_mult (iota gamma) (S i) hnil)).
             left; right; rewrite  <- hcons_mult_comm.
             apply hcons_mult_S1.
             left;  apply iota_succ_R1; auto.
             auto. 
        +  rewrite iota_tail; eauto with T1.
           * destruct a; simpl; destruct IHalpha2.
             eauto with T1. 
             intro H3; subst alpha2; discriminate.   
             destruct H2. 
             { 
               eexists; left.
               rewrite <- iota_iotas in H2.
               rewrite <- iota_iotas in H2.
               inversion H2.
               split.
               right;  apply hcons_mult_S0.
               inversion H3;  auto.
               apply nf_dS.
               eauto with T1.
               eauto with T1.
             }
             exists x.
             replace (hcons (iota alpha1) (hcons_mult (iota alpha1) n (iotas alpha2)))
             with
             (hy_app (hcons_mult (iota alpha1) (S n) hnil) (iotas alpha2)).
             replace (hcons (iota alpha1) (hcons_mult (iota alpha1) n
                                                  (iotas (dS i alpha2))))
             with
             (hy_app (hcons_mult (iota alpha1) (S n) hnil) (iotas (dS i alpha2))).
             right; apply R2_app.
             rewrite <- iota_iotas in H2.
             rewrite <- iota_iotas in H2.
             auto.
             apply nf_dS; eauto with T1. 
             eauto with T1. 
             cbn; f_equal;  clear H2 H1 H0  i0 H IHalpha1.
             induction n;simpl;auto.
             f_equal;  auto.
             clear H2 H1 H0 i0 H IHalpha1; induction n;simpl;auto.
             f_equal.
             auto.
           *  intro ; subst alpha2; destruct a; discriminate.
      - destruct s.
        destruct a.
        rewrite H2.
        rewrite dS_succ; auto. 
        apply iota_succ_round; auto. 
        Unshelve.
        exact 0.
    Qed.

    (** We are now able to find a contradiction in the hypotheses *)
    
    Let h := iota alphamax.
    Let beta_h := m h.
    Let h' := Iota beta_h.
    
    Remark R1 : lt  (proj1_sig beta_h)  alphamax.
    Proof.
      unfold beta_h; auto.
      destruct (m h).
      simpl. 
      destruct l.  destruct a. auto.
    Qed.

    (** We apply a result from Ketonen and Solovay.
    beta_h is reachable from alphamax by a sequence of
    dS n operations *)
    
   Remark R2 : {n:nat | dS_plus n alphamax  (proj1_sig beta_h)}.
    Proof. 
      apply L2_6_1; auto. 
      split.
      - destruct beta_h; simpl.
        destruct l;auto.
      - split.
        + apply R1.
        + apply alphamax_nf.
    Defined.


    Remark R3 : alphamax <> zero.
    Proof. 
      intro H; specialize (m_bounded (hyd1 head)).
      clear h' beta_h; subst; inversion 1.
      T1_inversion H.
    Qed.

    Remark R4 : forall n alpha beta, nf alpha -> nf beta -> alpha <> zero ->
                                     dS_rel n alpha beta ->
                                     iota alpha -+-> iota beta.
    Proof.
      inversion 4; subst; left; apply d_S_iota; auto.
    Qed.   

    Remark R5 : forall n alpha beta, nf alpha -> nf beta -> alpha <> zero ->
                                     dS_plus n alpha beta ->
                                     iota alpha -+-> iota beta.
    Proof.
      induction 4.
      -  apply R4 with n;auto.
      -  destruct (T1_eq_dec y zero).
         +  subst;  assert (z=zero).
            {
              clear H2 H1 H0 IHclos_trans_1n.
              assert
                (forall x a,
                   clos_trans_1n T1 (dS_rel n) x a -> x = zero -> a = zero).
              { induction 1.
                - intro ; subst; inversion H0;  reflexivity.
                -  intros; apply IHclos_trans_1n.
                   subst; inversion H0; reflexivity.
              }
              eapply H0; eauto.
            }
            subst z; left; inversion H2.
            rewrite H5;  apply d_S_iota;   auto. 
         + constructor 2 with (iota y).
           * inversion H2; apply d_S_iota;  auto. 
           * apply IHclos_trans_1n; auto. 
             inversion H2.
             apply nf_dS; auto. 
    Qed.

    Corollary  C5 : forall alpha beta,  alpha < beta ->
                                        iota beta -+-> iota alpha.
    Proof. 
      intros alpha beta H.
      assert (H0: nf beta) by eauto with T1.
      destruct (L2_6_1 H0 H) as [n H1].
      eapply R5; eauto with T1.
      intro;subst beta; apply (not_LT_zero H).
    Qed. 


    Remark R6 : h -+-> h'.
    Proof. 
      unfold h, h';  destruct R2.  unfold Iota.
      destruct beta_h. apply R5 with x; auto.
      destruct l; auto.
      apply R3.
    Qed.

    Corollary mh'_lt_mh : Lt (m h') (m h).
    Proof. apply m_variant_Lt; apply R6. Qed.

    Remark R8 : forall h, nf (m_extract h).
    Proof.
      intro h0.  unfold m_extract. destruct (m h0).
      destruct l;auto. 
    Qed.

    (** With a transfinite induction,
     we prove the second inequality for getting a contradiction.

     In the case where alpha is a limit ordinal, we use the fact that
     alpha is the least upper bound of all the (d i alpha). 
 
     *)

    Lemma m_ge : forall alpha, nf alpha -> alpha <= m_extract (iota alpha).
    Proof.
      intros alpha ; transfinite_induction alpha.
      clear alpha; intros alpha Hrec H.
      destruct (T1_eq_dec alpha zero).
      + subst;  split;auto with T1.
        split.
        auto with T1.
        apply R8.
      +   destruct (zero_succ_limit_dec H).
          * destruct s.
            { subst ; now destruct n. }
            { eapply limit_strict_lub with (s := fun i => dS i alpha).
              apply dS_limit; auto. 
              intro i0;  assert (iota alpha -1-> iota (dS i0 alpha)). 
              {   apply d_S_iota;  auto. }
              apply LE_LT_trans with (m_extract (iota (dS i0 alpha))); auto.
              { apply Hrec; auto. 
                apply nf_dS;  auto.
                destruct (dS_limit H); auto. 
                destruct (H1 i0).
                tauto.
                eapply nf_dS;auto.
              }
              destruct V.
              specialize (decr _ _ H0).              

              unfold m_extract.
              unfold Lt in decr.
              split.
              destruct (m (iota (dS i0 alpha))).
              destruct l;auto.
              split.
              now destruct (m (iota (dS i0 alpha))),  (m (iota alpha)).
              destruct (m (iota alpha)).
              destruct l;auto.
            }
            
          *  destruct s.
             destruct a; subst.
             assert (x < m_extract (iota (succ x))).
             { 
               apply LE_LT_trans with (m_extract (iota x)).
               apply Hrec; auto. 
               apply lt_succ; auto.
               split.
               unfold m_extract.
               destruct (m (iota x)).
               destruct l;auto.

               split. 
               unfold m_extract.
               generalize (iota_succ_round x H0 ).
               intro H1.
               destruct V  as [wf decr]; specialize (decr _ _ H1).
               red in decr.
               destruct (m (iota x)) , (m (iota (succ x))). auto.
               unfold m_extract.
               destruct (m (iota (succ x))).
               destruct l;auto.
             }
             apply  LT_succ_LE; auto. 
    Qed. 


    Let Le := clos_refl _ Lt.


    Lemma Lt_lt : forall x y (H: x < alphamax)(H0:y < alphamax),
                    Lt (exist (fun x =>  x < alphamax) x H)
                       (exist (fun x =>  x < alphamax) y H0) <->
                    lt x y.
    Proof. unfold Lt; intros; tauto. Qed.

    Lemma Le_le : forall x y (H: x < alphamax)(H0:y < alphamax),
                    Le (exist (fun x =>  x < alphamax) x H)
                       (exist (fun x =>  x < alphamax) y H0) <->
                    le x y.
    Proof.
      unfold Le; intros.
      split.
      - inversion  1.
        + subst; rewrite Lt_lt in H2.
          now apply lt_le_incl.
        + subst; auto with T1.
      - 
        intro H1; destruct (le_lt_or_eq _ _ H1).
        + subst;  assert (H0 = H).
          { apply LT_proof_unicity. }
          subst; right.
        + left; rewrite  Lt_lt; auto. 
    Qed. 

    
    Lemma m_ge_2 : forall alpha:t, Le alpha (m (Iota alpha)).
    Proof.
      destruct alpha as [x l];  unfold Iota;   assert (nf x).    
      { destruct l; auto. }
      generalize (m_ge x H); intro H0.
      unfold m_extract in H0; cbn.
      destruct (m (iota x)) as [x0 Hx0].
      rewrite Le_le; destruct H0;  tauto. 
    Qed. 

  End Assumed_Variant.

  (** The relation Lt on type t is too simple for defining a variant *)
  
  Instance alpha_too_small : TooSimple t Lt.
  Proof.
    exists Iota (iota alphamax).
    - intros;  now apply m_ge_2.
    - intros m V; unfold Iota;   generalize (R6 m V).
      destruct (m (iota alphamax));  now   unfold Iota.
  Qed.
  
End Epsilon0_needed.




