(** Pierre Castéran, University of Bordeaux and LaBRI *)


Require Import Hydra_Games Simple_LexProd Relations Relation_Operators
Arith Peano_dec Omega WfVariant TooSimpleVariant.

(** There is no measure into omega^2  for proving termination
of all hydra battles *)


Section nat2_too_simple.

   (** Let us assume there is a variant from [Hydra] into nat*nat (with the
   lexicographic oordering)  for proving the   termination of all hydra battles *)
  Let t := (nat * nat)%type. 

  
  Let h := hyd1 (hyd2 head head).
  (** To every pair (i,j) of natural numbers we associate an hydra 
        with i branches of length 2 and j branches of lenght 1 *)

  Let iota (p: t) := 
    node (hcons_mult (hyd1 head) (fst p)
                     (hcons_mult head (snd p) hnil)).

  

  (** non-dependent lexicographic strict ordering on nat*nat *)
  Let lt2 : relation t := lexico Peano.lt Peano.lt.
  Infix "<" := lt2.
  
  (** reflexive closure of lt2 *)
  Let le2 := clos_refl _ lt2.
  Infix "<=" := le2.

  Section Assumed_Variant.

    Variable m : Hydra -> t.
    Context  (V : WfVariant round lt2 m).

    Hint Resolve (decr V).
    Let h' := iota (m h).

    
 

 

 
   (** ** Properties of lt2 and le2 
    *)
   Lemma le2_introl :
     forall i j k:nat, (j <= k)%nat -> (i,j) <= (i,k).
   Proof.
     intros i j k H; destruct (Lt.le_lt_or_eq j k H).
     - left; now right.
     - subst;now right.
   Qed.

   Lemma le2_0 : forall p: t,  (0,0) <= p.
   Proof.
     destruct p, n.
     - apply le2_introl;  auto with arith.
     -  left; left;  auto with arith.
   Qed.

   Lemma le2_1 : forall i j p,  (i,j) < p -> (i, S j) <= p.
   Proof.
     intros i j p H; destruct p as (k,l).
    -  inversion_clear H.
     + left; now left.
     + destruct (Lt.le_lt_or_eq _ _ H0).
      * left;right;auto.
      * subst l; now right.
   Qed.
   
   Lemma le2_2 : forall i p, (forall j, (i,j) < p) ->
                             (S i, 0) <= p.
   Proof.
     intros i p H ;  destruct p as (k,l).
     destruct (Nat.eq_dec (S i) k).
     - subst k.
       destruct l.
       +  now right.
       +  left;  right;  auto with arith.
     -   generalize (H (S l)); intro H'.
         inversion_clear H'.
         destruct l.
         +  destruct (Lt.le_lt_or_eq _ _ H0).
            *  left; now left.
            *  subst k; now right.
         + left; left; omega.
         +  now destruct (Nat.nlt_succ_diag_l l).
   Qed.

   Instance Sto_lt2 : StrictOrder lt2.
   Proof. apply Strict_lex; apply Nat.lt_strorder.
   Qed.

   Lemma le_lt2_trans : forall p q r, p <= q -> q < r -> p < r.
   Proof.
     destruct p, q, r; inversion_clear 1; auto.
     intros; transitivity (n1,n2);auto.
   Qed.   


   
   (** Let us come back to the properties of the variant m *)

   Lemma m_strict_mono : forall h h',
                                h -+-> h' ->  m h' < m h.
   Proof.
     induction 1;auto.
     transitivity (m y);auto.
   Qed.
   
  Lemma mh_not_null : m h <> (0,0).
  Proof.
   intro H; pose (h1 := hyd1 (hyd1 head)).
   assert (first_round : h -1-> h1).
   {
     exists 0; right; left; constructor 1 with (n:=0).
     split; left.
   }
   specialize (decr V  _ _ first_round).
   rewrite H .
   inversion_clear 1; omega.
  Qed. 


  Lemma iota_limit : forall i j, round_n j (iota (S i,0)) (iota (i, S j)).
  Proof.
    intros i j;    induction i.
    - unfold iota;simpl.
     right;  left.
      change (hcons head (hcons_mult head j hnil)) with
      (hcons_mult head (S j) hnil).
      left;  split;    left.
    -  right;  left; cbn. 
       rewrite <- hcons_mult_comm;  right. 
       apply hcons_mult_S1.
       left; split; constructor. 
  Qed.

   
  Lemma iota_succ : forall n i j, round_n n (iota (i, S j))  (iota (i, j)).
  Proof.
    unfold iota; cbn; left;split.
    apply hcons_mult_S0;  constructor.
  Qed.

  Lemma iota_lt_r : forall i j k,   (j < k)%nat ->
                                    iota (i, k) -+-> iota (i,j).
  Proof.
   induction 1.   
   - left;   exists 1; apply iota_succ.
   -   apply round_plus_trans with   (iota (i,  m0)).
       + left;   exists 1; apply iota_succ.
       + assumption. 
  Qed.

  Lemma iota_limit2 : forall i j, iota (S i,0) -+-> iota (i,  j).
  Proof.
    intros i j. 
    apply round_plus_trans with (iota (i, S j)).
    left;exists j.  apply iota_limit.
    left;exists 0; apply iota_succ.
   Qed. 
           
  Lemma iota_lt_l : forall i j k l,   (i < j)%nat ->
                                    iota (j, k) -+-> iota (i,l).
  Proof.
   induction 1.
   -  destruct k.
      apply iota_limit2.
      apply round_plus_trans with (iota (S i, 0)). 
       apply iota_lt_r; auto with arith. 
apply     iota_limit2.
   -  destruct k. 
 apply round_plus_trans with (iota (m0, 0)).
    
     apply iota_limit2.
  auto. 
    apply round_plus_trans with (iota (S m0, 0)).
     apply iota_lt_r; auto with arith.
    apply round_plus_trans with (iota (m0, S k)).
      apply iota_limit2.
 auto. 
Qed. 



     Lemma round_0 : h -+-> h'. 
  Proof.
    unfold h'; case_eq  (m h); intros i j Hj.  
    destruct i.
    
    (*  i = 0 *) 
    - unfold h; right with (hyd1 (hyd1 head)).
      exists 0;right;  left.
      left with (n:=0).
      split; left.
      destruct j.
      now destruct mh_not_null.
      unfold iota;cbn;left.
      exists j;right; left; constructor 1;  split;  left.

    (* i > 0 *)
      
    - unfold iota; cbn;  destruct j.
      +   left; unfold iota, h; cbn.
          exists i; right; left; left . 
          split; left.
      +  right with (iota (S (S i), 0)). 
         unfold iota;  simpl fst ; simpl snd.
         cbn; exists (S i); right.
        left; left;  split; left.
        unfold iota; cbn; left.
        exists j; right; left;   right. 
        rewrite <- hcons_mult_comm;  apply hcons_mult_S1.
        left;   split;  left.
  Qed.


  Corollary mh'_lt_mh : m h' < m h.
  Proof.
    apply m_strict_mono,  round_0.
  Qed.


  Lemma compat : forall p q: t,  q < p -> iota p -+-> iota q.
  Proof. 
    inversion 1.
    destruct p, q.
    -    subst.
     now apply iota_lt_l.
    - apply iota_lt_r;    auto.     
  Qed. 




  (** We prove now a lemma similar to Omega_Small.m_ge. 
      Notice the use of well-founded induction *)
  
  Lemma m_ge : forall p : t,   p <= m (iota p).
  Proof.
    intro p ; pattern p;
    apply  well_founded_induction with (R := lt2).
    { apply wf_lexico;   apply lt_wf. }
    intros (i,j) IHij.
    destruct j.
    - destruct i. 
     +  (* p = (0,0) *)
        apply le2_0. 
     +  (* p = (S i, 0) *)
       apply le2_2;    intro j; transitivity (i, S j).       
      * right;auto.
      * apply le_lt2_trans with (m (iota (i, S j))).
         { apply IHij; left;auto. }
         { apply (decr V).
           exists j; apply iota_limit. }
    -   specialize  (IHij (i,j)).
        assert ((i, j) < (i, S j)).
        { right;auto with arith. }
        apply le2_1.
        apply le_lt2_trans with (m (iota (i, j))); auto.
        + apply (decr V); exists 0; apply iota_succ.
  Qed.

  End Assumed_Variant.

  
  Instance omega2_too_small : TooSimple t lt2.
  Proof.
    exists iota h.
    -  intros; now apply m_ge.
    - intros;  now apply round_0.     
  Qed.
  
  
End nat2_too_simple.

