(* Pierre Casteran, LaBRI,  University of  Bordeaux *)

(** *

 We want to adapt Schutte's definition of critical ordinals :
   
   Cr(zero) = AP   (the set of additive principal ordnals )
  
   if zero < alpha, then
   Cr(alpha) is the set of ordinals x 
       such that for all beta<alpha, x is a fixpoint of the
       ordering function of Cr(beta)
*) 


Require Import Arith.
Require Import Epsilon.
Require Import Ensembles.
Require Import Schutte.
Require Import Ordering_Functions.
Require Import PartialFun.
Require Import Denumerable.
Require Import Plus.

Require Import AP.
Require Import Classical.
Require Import CNF.
Require Import PartialFix.
Require Import Well_Orders.
Require Import  ProofIrrelevance.





Set Implicit Arguments.


(* let us define a functional, the fixpoint of which we shall consider *)


Definition critical_fun : forall alpha : OT,
       (forall beta : OT, beta < alpha -> Ensemble OT) ->
       ordinal alpha -> OT -> Prop
:= 
   fun (alpha :OT)
        (Cr : forall beta, 
                beta < alpha -> 
               Ensemble OT) 
       (o_alpha :ordinal alpha)
       (x : OT)   => ordinal x /\(
            (alpha = zero /\ AP x) \/
             (zero < alpha /\
              forall beta (H:beta < alpha), Cr beta H x /\
                 the_ordering_function (Cr beta H) x =x)).

Definition critical : forall alpha, (ordinal alpha) -> Ensemble OT :=
    (PFix ordinal lt_ordinal all_ord_acc 
          (fun (_:OT) => Ensemble OT) critical_fun).

Definition Cr (alpha : OT) : Ensemble OT :=
   fun beta => 
       exists H: ordinal alpha, critical  H beta.

Definition phi : OT-> OT -> OT 
    := fun alpha =>  the_ordering_function (Cr alpha).

Definition epsilon0 := phi (F 1) zero.



Lemma critical_extensional : forall (x:OT) 
          (f g : forall y : OT, y < x -> (fun _ : OT => Ensemble OT) y)
          (Ox : ordinal x),
         (forall (y : OT) (p : y < x), f y p = g y p) ->
          ((critical_fun  f Ox :Ensemble OT)=
            (critical_fun  g Ox:Ensemble OT)).
Proof.
 intros x f g H H0.  
 apply Extensionality_Ensembles.
 split.
 unfold critical_fun;red.
 unfold In;intros.
 case H1;auto.
 intros.
 split.
 auto.
 case H3.
 auto.
 right.
 split.
 case H4;auto.
 intros.
 decompose [and] H4.
 rewrite <- (H0 beta H5).
 auto.
 unfold critical_fun;red.
 unfold In;intros.
 case H1;auto.
 intros.
 split.
 auto.
 case H3.
 auto.
 right.
 split.
 case H4;auto.
 intros.
 decompose [and] H4.
 
 rewrite (H0 beta H5).
 case (H7 beta H5).
 
 auto.
Qed.

Arguments critical [alpha].


Lemma critical_eq 
     : forall (alpha : OT) (H : ordinal alpha),
       critical  H =
       critical_fun 
         (fun (y : OT) (h : y < alpha) =>
          critical  (lt_ordinal   h))
         H.
Proof  (PFix_eq  ordinal lt_ordinal all_ord_acc
 (fun _ : OT => Ensemble OT) critical_fun critical_extensional
).

(* critical elimination rule *)

Lemma critical_e : forall alpha (Halpha: ordinal alpha),
                forall x,
                   critical Halpha x -> ordinal x /\(
                   (((alpha = zero) /\ (critical  Halpha x <-> AP x)) \/
                   (zero < alpha /\
                                  (forall beta (H:beta < alpha),
                                         critical (lt_ordinal H) x /\
                                       the_ordering_function 
                                         (critical  (lt_ordinal H)) x =x)))).
Proof.
 intros alpha Halpha.
 intro x.
 rewrite (critical_eq  Halpha).
 unfold critical_fun at 1.
 intros.
 split.
 case H;auto.
 decompose [and] H.
 
 case H1;intros.
 clear H1.
  left.
split.
 case H2;auto.
 split.
 intro.
 case H2;auto.
 intro.
 red.
 split.
 auto.
 left;auto.
 right.
 auto.
Qed.
 

(* introduction rule *)

Lemma critical_ord : forall alpha (H:ordinal alpha) x,
                            critical H x -> ordinal x.
Proof.
 intros.
 case (critical_e H x).
 auto.
 auto.
Qed.


Lemma inh_ens_OT : inhabited (Ensemble OT).
Proof.
 split.  exact (fun a : OT => False).
Qed.



(* 
Inductive and_dep (P : Prop)(Q : P -> Prop) :Prop :=
conj_dep : forall (p:P), Q p -> and_dep  Q.

Implicit Arguments and_dep [].

Check (critical :  forall alpha : OT, ordinal alpha -> Ensemble OT).

*)

(*

critical
     : forall alpha : M ON, ordinal alpha -> Ensemble OT
*)


(* Cr alpha is a dependent conjunction of 
   H: ordinal alpha and being critical *)



Lemma critical_zero : forall x (H:ordinal zero) , AP x -> critical H x.
Proof.
 intros.
 rewrite critical_eq. 
 red.
  split.
  case H0;eauto with schutte.
 
left;auto.
Qed.


Lemma Cr_zero :  forall x, AP x -> Cr zero x.
Proof.
 intros.
 exists ordinal_zero.
 apply critical_zero.
 auto.
Qed.



Lemma critical_pos : forall alpha (Oalpha:ordinal alpha),
   zero < alpha ->
    forall x, ordinal x ->
   (forall beta (H:beta < alpha), (critical  (lt_ordinal  H) x) /\
         the_ordering_function (critical  (lt_ordinal  H)) x =x) ->
   critical Oalpha x.
 intros.
 rewrite critical_eq.
 red.
  split.
 auto.
right.
 split;auto.
Qed.


Lemma Cr_rw : forall alpha (H:ordinal alpha), Cr alpha = critical H.
Proof.
 intros.
 apply Extensionality_Ensembles.
 unfold Cr.
 split.
 intros beta Hbeta.
 red in Hbeta.
 case Hbeta.
 intro p.

 rewrite (proof_irrelevance( ordinal alpha) H p).
 auto.
 red.
 intros x Hx.
 generalize (critical_ord H x Hx).
 intro.
 red.
 exists H.
 auto.
Qed.


Lemma Cr_pos : forall alpha, zero < alpha ->
                      forall x, ordinal x ->
   (forall beta ,beta < alpha -> (Cr beta x) /\
         the_ordering_function (Cr beta) x =x) ->
   Cr alpha x.  
Proof.
 intros.
 assert (ordinal alpha).
 ordi.
 exists H2.
 apply critical_pos;auto.
 intros.
 case (H1 beta H3);intros.
 split;auto.
 rewrite <- Cr_rw;auto.
  rewrite <- Cr_rw;auto.
Qed.


 
 
Lemma critical_zero_inv : forall x, critical ordinal_zero x -> AP x.
Proof.
 intros.
 case (critical_e ordinal_zero x H).
 destruct 2.
 
tauto.
 case H1.
 intros.
 case (@lt_irr zero).
 auto.
Qed.

Lemma Cr_zero_AP :  Cr zero = AP.
Proof.
 apply Extensionality_Ensembles.
 rewrite (Cr_rw  ordinal_zero).
 split.
 red;intros;red;apply critical_zero_inv;auto.
 red;intros.
 red.
 apply critical_zero;auto.
Qed.


 



Lemma critical_pos_inv : forall alpha (Oalpha:ordinal alpha),
   zero < alpha ->
    forall x, 
   critical Oalpha x ->
   (forall beta (H:beta < alpha), (critical  (lt_ordinal  H)) x /\
         the_ordering_function (critical  (lt_ordinal H)) x =x).

 intros.
 case (critical_e Oalpha x H0).
 destruct 2.
 decompose [and] H3.
subst alpha.


 case (@not_lt_zero beta). 
 auto. 
 case H3.
 intros.
  auto.

Qed.



Lemma critical_lt : forall alpha beta (H:ordinal alpha)
                                      (H0:ordinal beta),
   beta < alpha -> forall x, critical H x -> critical H0 x.
Proof.
 intros.
 case (critical_e  H x);auto.
 intros Hx [[H3 H4]| [H3 H4]].
 subst alpha.
 case (@not_lt_zero beta);auto.
 case (H4 beta H1).
 intros.
 rewrite (proof_irrelevance( ordinal beta) H0 (lt_ordinal  H1)).
 auto.
Qed.



Lemma Critical_incl : forall alpha beta  (H :beta <= alpha)
    (Ha : ordinal alpha)(Hb : ordinal beta),
    Included (critical Ha)  (critical Hb).
Proof.
 intros.
 case (le_disj H).
 
 intro;subst beta.
rewrite (proof_irrelevance _ Ha Hb).
 auto.
 intro.
 assert (zero < alpha).
 eauto with schutte.
 red;unfold In;intro.
 intro.
 generalize (critical_pos_inv  (gt_ordinal  H1) H1 x ).
 intros.
 rewrite (proof_irrelevance _ Hb (lt_ordinal   H0)).
rewrite  (proof_irrelevance _ Ha (gt_ordinal  H1)) in H2.
generalize (H3 H2).
 intro.
 generalize (H4 _ H0).
 intro.
 case H5;auto.
Qed.



Lemma Cr_incl : forall alpha beta ,beta <= alpha ->
    Included (Cr alpha) (Cr beta).
Proof.
 intros.
  rewrite (Cr_rw  (ge_ordinal H)).
 rewrite (Cr_rw  (le_ordinal H)).
 
 apply Critical_incl;auto.
Qed.


Ltac unfold_CR  alpha := let H := fresh in
                      (assert(H: ordinal alpha);[try ordi | rewrite (Cr_rw  H)]).

Lemma Cr_incl' : forall alpha beta ,beta <= alpha ->
    Included (Cr alpha) (Cr beta).
Proof.
 intros.
 unfold_CR alpha.
 unfold_CR beta.
 apply Critical_incl;auto.
Qed.




Lemma phi0_well_named : forall alpha, phi0 alpha = phi zero alpha.
Proof.
 intro alpha;unfold phi.
 rewrite Cr_zero_AP.
 reflexivity.
Qed.

Lemma CR_1_eq : forall x, ordinal x -> (Cr (F 1) x <->
                                       AP x /\ phi0 x = x).
Proof.
 intros.
 unfold_CR (F 1).
 generalize (critical_pos_inv  H0).
 intros.
 split.
 intros.
 assert (zero < F 1).
 simpl.
 apply lt_succ.
 auto with schutte.
 generalize (H1 H3 x H2).
 intros.
 generalize (H4 zero H3).
 intros (H5,H6).
 split.
 apply critical_zero_inv.
 rewrite (proof_irrelevance _ (lt_ordinal  H3) ordinal_zero) in H5.
 auto.
 rewrite (proof_irrelevance _ (lt_ordinal H3) ordinal_zero) in H6.
 unfold phi0.
 replace AP with (critical ordinal_zero).
 auto.
 apply Extensionality_Ensembles.
 split.
 red.
 red.
 intros.
 apply critical_zero_inv.
 auto.
 red.
 red.
 intros;apply critical_zero.
 auto.
 intros (H2,H3).
 clear H1.
 rewrite <- Cr_rw.
 apply Cr_pos.
 simpl.
 apply lt_succ.
 auto with schutte.
 auto.
 intros.
 assert (beta=zero).
 simpl in H1.
 generalize (lt_succ_le_2 ordinal_zero H1).
  intro H4;rewrite (le_alpha_zero H4).
 auto.
 subst beta.
 split.
 rewrite Cr_zero_AP;auto.
 rewrite Cr_zero_AP;auto. 
Qed.










 
