
(** * Compatibility with StdLib exponentiation functions  

Function for computing powers already  exist in %\coq%'s standard library. We provide
equivalence theorems with functions defined in our module [Pow].


 *)
(* begin hide *)
Require Import Monoid_def Omega Monoid_instances
     ArithRing  Pow.
(* end hide *)

(** ** really logarithmic versions of N.pow, Pos.pow and Z.pow 

We propose three functions that are extensionnaly equivalent to functions
of the standard library.
These functions are defined in the %\href{../Powers/SRC/Pow.v}{Powers/SRC/Pow.v}% module;  we just rename them for readability's sake.

*)

Definition N_pow (a n: N) := N_bpow a n.

Definition Pos_pow (a n : positive) := Pos_bpow a  n.

Definition Z_pow (x y : Z) :=
 match y with 
| 0%Z => 1%Z
| Z.pos p => N_bpow x (Npos  p)
| Z.neg _ => 0%Z
end.

(** **  Equivalence between S.L. definitions and ours *)

Section Equivalence.
Variable (A: Type)
         (op : Binop A)
         (one : A).

Context (M : Monoid  op one).
Open Scope M_scope.

Check fun x y:A =>  x * y.

(* begin hide *)
Ltac monoid_rw :=
    rewrite (@one_left A op one M) || 
    rewrite (@one_right A op one M)|| 
    rewrite (@op_assoc A op one M).

  Ltac monoid_simpl := repeat monoid_rw.

 Ltac power_simpl := repeat (monoid_rw || rewrite <- power_of_plus).

(* end hide *)

 Let pos_iter_M x := Pos.iter (op x).

 (**
  
  During an execution of the binary exponentiation algorithm
  through the function [binary_power_mult],
  the  "accumulator" [acc] is always a power of $x$. Thus, even if the 
  considered monoid is not abelian, the accumulator commutes with any 
  other power of $x$.
 

 *)
 Let is_power_of (x acc:A) := exists i, acc = x ^ i.


Lemma  Pos_iter_op_ok_0 : 
  forall p x acc, is_power_of x acc ->
                  pos_iter_M x acc  p = binary_power_mult op x acc p .
(* begin hide *)
Proof.
  induction p.
  -  simpl.
     intros;  rewrite IHp.
     rewrite (IHp x acc).
     repeat rewrite (binary_power_mult_ok M).
     rewrite <- (sqr_def M x).
     rewrite power_of_power.
     rewrite (mult_comm  (Pos.to_nat p) 2)%nat. 
     destruct H as [i e].
     repeat rewrite  (op_assoc (Monoid:=M)). 
     subst acc.
     rewrite  <- (power_commute_with_x ).
     replace (2* Pos.to_nat p)%nat with (Pos.to_nat p + Pos.to_nat p)%nat.

     + rewrite power_of_plus.
       now rewrite (op_assoc (Monoid:=M)). 
     + ring.
     + assumption.
     +  rewrite IHp;auto.
        rewrite (binary_power_mult_ok M).
        destruct H as [i Hi].
        exists (i + Pos.to_nat p)%nat.
        rewrite power_of_plus.
        subst acc;auto.

  - simpl;  intros;   rewrite IHp.
    rewrite (IHp  x acc).
    repeat rewrite (binary_power_mult_ok M); rewrite  <- (sqr_def M).
    repeat (rewrite (op_assoc (Monoid:=M))).
    rewrite power_of_power.
    rewrite (mult_comm  (Pos.to_nat p) 2)%nat. 
    replace (2* Pos.to_nat p)%nat with (Pos.to_nat p + Pos.to_nat p)%nat.  
    + rewrite power_of_plus;now  rewrite op_assoc.
    + ring.
    + assumption.
    + rewrite IHp, (binary_power_mult_ok M).
      destruct H as [i Hi].
      exists (i + Pos.to_nat p)%nat.
      rewrite power_of_plus.
      subst acc;auto.
      auto.
  -   intros; simpl. change (op x acc) with (x * acc). 
      destruct H as [i Hi]; subst acc.
      apply power_commute_with_x.
Qed.
(* end hide *)

Lemma  Pos_iter_op_ok:
  forall p x,  
    pos_iter_M x one  p = binary_power_mult  op x one p .
(* begin hide *)
Proof.
 intros p x; rewrite Pos_iter_op_ok_0;auto.
 now exists 0%nat.
Qed.
(* end hide *)

Lemma  Pos_iter_ok: forall p x,  N_bpow x (Npos p) = pos_iter_M x one   p.
(* begin hide *)
Proof.
  intros p x; rewrite Pos_iter_op_ok_0.
  unfold N_bpow.   
  rewrite Pos_bpow_ok.
 generalize  (binary_power_mult_ok M (E_eq := eq_equiv)).
 unfold equiv, eq_equiv.
intro H; rewrite H. 
now      rewrite (one_left (Monoid := M)).
 exists 0%nat.
 reflexivity.
Qed.
(* end hide *)

End Equivalence.

Lemma Pos_pow_power : forall n a : positive ,
                        (a  ^ n )%positive  = power a  (Pos.to_nat n).
(* begin hide *)
Proof.
 intros n a; unfold Pos.pow.
 rewrite (Pos_iter_op_ok _ _ _ PMult).
 generalize  (binary_power_mult_ok  (E_one := xH) (E_eq := eq_equiv)).
 intro H; rewrite H. 
 rewrite one_left.
 reflexivity.
Qed.
(* end hide *)

Lemma Npos_power_compat : forall (n:nat)(a:positive),
                            Npos (power a  n) = power (Npos a)  n.
(* begin hide *)
Proof. 
  induction n.
 - now compute.
 -  intros;  repeat rewrite  power_eq2; now rewrite <- IHn.
Qed.
(* end hide *)

Lemma N_pow_power : forall n a , (a  ^ n )%N  = power a (N.to_nat n).
(* begin hide *)
Proof.
 intros n a; unfold N.pow.
 destruct n as [ | p].
-   compute; reflexivity.
-   simpl; destruct a.
  +   assert (0 <> Pos.to_nat p)%nat.
      { generalize (Pos2Nat.is_pos p).
        intro;omega.
      }
      generalize H;  clear H; generalize (Pos.to_nat p).
       intro n;induction n;simpl;auto.
       destruct 1;auto.
   + rewrite Pos_pow_power;now rewrite Npos_power_compat.
Qed.
(* end hide *)



Lemma N_pow_compat : forall n (a:N),   (a  ^ n )%N = N_pow a n.
Proof. 
 intros. unfold N_pow;  rewrite N_bpow_ok; now rewrite  N_pow_power.
Qed.


Lemma Pos_pow_compat : forall n (a:positive),  (a  ^ n )%positive = Pos_pow  a n.
Proof.
 unfold Pos_pow; intros;rewrite Pos_bpow_ok; now rewrite  Pos_pow_power.
Qed.


Lemma nat_power_ok : forall a b:nat, (a ^ b)%nat = (a ^ b)%M.
induction b;simpl.
 trivial.
  rewrite IHb;auto.
Qed.

  Lemma Pos2Nat_morph : forall (n:nat)(a : positive), 
     Pos.to_nat (a ^ n)%M = Pos.to_nat a ^ n.
  induction n.
  intro;simpl; reflexivity.    
  intro; simpl.
   rewrite Pos2Nat.inj_mul.
   rewrite IHn.
    reflexivity.
  Qed.

Lemma Z_pow_compat_pos : forall (p:positive) (x:Z), 
        Z.pow_pos x p = N_bpow x (Npos  p).
(* begin hide *)
Proof. 
 intros; unfold Z.pow_pos.
 rewrite (Pos_iter_op_ok _ _ _ ZMult).
  rewrite (binary_power_mult_ok (E_one := 1%Z) (E_eq := eq_equiv)).
 rewrite N_bpow_ok.
 rewrite (one_left  (Monoid := ZMult)) . 
 reflexivity.
Qed.
(* end hide *)



Lemma Z_pow_compat : forall x y: Z, Z_pow x y = (x ^ y)%Z.
(* begin hide *)
Proof.
 intros x y; unfold Z_pow, Z.pow.
 destruct y; [ | rewrite Z_pow_compat_pos | ]; reflexivity.
Qed.
(* end hide *)

(** ** Tests 

Let us chose a big exponent, and a computation that does not create 
big numbers. So, the following computation time will be proportional to
 the number of recursive calls, hence to the number of multiplications

*)

(** Time consuming !
[[
Time Compute (1 ^ 56666667)%Z.
]]
*)

(** 
[[
Finished transaction in 10.195 secs (10.19u,0.004s) (successful)
]]
*)

Time Compute (Z_pow 1 56666667)%Z.
(** 
[[
Finished transaction in 0. secs (0.u,0.s) (successful)
]]
*)


(** *** Adapting lemmas from Standard Library 

Standard library already contains many lemmas about 
[N.pow], [Pos.pow] and [Z.pow]. By using our extensional equivalence properties, one can 
  easily prove the same properties for our implementation of the same functions.

We just give a simple example of this adaptation.

*)
Section Adaptation.
Local Notation "x ^ y" := (N_pow x y):N_scope.

Lemma N_size_gt : forall n : N, (n < 2 ^ N.size n)%N.
Proof. 
  intro n; rewrite  <- N_pow_compat.
  apply N.size_gt.  
Qed.

End Adaptation.
 


